@@ -417,7 +417,7 @@ PHP_FUNCTION(posix_ctermid)
 /* }}} */
 
 /* Checks if the provides resource is a stream and if it provides a file descriptor */
-static int php_posix_stream_get_fd(zval *zfp, int *fd) /* {{{ */
+static int php_posix_stream_get_fd(zval *zfp, zend_long *fd) /* {{{ */
 {
 	php_stream *stream;
 
@@ -426,10 +426,15 @@ static int php_posix_stream_get_fd(zval *zfp, int *fd) /* {{{ */
 	if (stream == NULL) {
 		return 0;
 	}
-	if (php_stream_can_cast(stream, PHP_STREAM_AS_FD_FOR_SELECT) == SUCCESS) {
-		php_stream_cast(stream, PHP_STREAM_AS_FD_FOR_SELECT, (void*)fd, 0);
-	} else if (php_stream_can_cast(stream, PHP_STREAM_AS_FD) == SUCCESS) {
-		php_stream_cast(stream, PHP_STREAM_AS_FD, (void*)fd, 0);
+
+	/* get the fd.
+	 * NB: Most other code will NOT use the PHP_STREAM_CAST_INTERNAL flag when casting.
+	 * It is only used here so that the buffered data warning is not displayed.
+	 */
+	if (php_stream_can_cast(stream, PHP_STREAM_AS_FD_FOR_SELECT | PHP_STREAM_CAST_INTERNAL) == SUCCESS) {
+		php_stream_cast(stream, PHP_STREAM_AS_FD_FOR_SELECT | PHP_STREAM_CAST_INTERNAL, (void*)fd, 0);
+	} else if (php_stream_can_cast(stream, PHP_STREAM_AS_FD | PHP_STREAM_CAST_INTERNAL) == SUCCESS) {
+		php_stream_cast(stream, PHP_STREAM_AS_FD | PHP_STREAM_CAST_INTERNAL, (void*)fd, 0);
 	} else {
 		php_error_docref(NULL, E_WARNING, "Could not use stream of type '%s'",
 				stream->ops->label);
@@ -444,7 +449,7 @@ PHP_FUNCTION(posix_ttyname)
 {
 	zval *z_fd;
 	char *p;
-	int fd;
+	zend_long fd = 0;
 #if defined(ZTS) && defined(HAVE_TTYNAME_R) && defined(_SC_TTY_NAME_MAX)
 	zend_long buflen;
 	int err;
@@ -454,14 +459,21 @@ PHP_FUNCTION(posix_ttyname)
 		Z_PARAM_ZVAL(z_fd)
 	ZEND_PARSE_PARAMETERS_END();
 
-	switch (Z_TYPE_P(z_fd)) {
-		case IS_RESOURCE:
-			if (!php_posix_stream_get_fd(z_fd, &fd)) {
-				RETURN_FALSE;
-			}
-			break;
-		default:
+	if (Z_TYPE_P(z_fd) == IS_RESOURCE) {
+		if (!php_posix_stream_get_fd(z_fd, &fd)) {
+			RETURN_FALSE;
+		}
+	} else {
+		if (!zend_parse_arg_long(z_fd, &fd, /* is_null */ NULL, /* check_null */ false, /* arg_num */ 1)) {
+			php_error_docref(NULL, E_WARNING, "Argument #1 ($file_descriptor) must be of type int|resource, %s given",
+				zend_zval_value_name(z_fd));
 			fd = zval_get_long(z_fd);
+		}
+		/* fd must fit in an int and be positive */
+		if (fd < 0 || fd > INT_MAX) {
+			php_error_docref(NULL, E_WARNING, "Argument #1 ($file_descriptor) must be between 0 and %d", INT_MAX);
+			RETURN_FALSE;
+		}
 	}
 #if defined(ZTS) && defined(HAVE_TTYNAME_R) && defined(_SC_TTY_NAME_MAX)
 	buflen = sysconf(_SC_TTY_NAME_MAX);
@@ -502,22 +514,28 @@ try_again:
 PHP_FUNCTION(posix_isatty)
 {
 	zval *z_fd;
-	int fd;
+	zend_long fd = 0;
 
 	ZEND_PARSE_PARAMETERS_START(1, 1)
 		Z_PARAM_ZVAL(z_fd)
 	ZEND_PARSE_PARAMETERS_END();
 
-	switch (Z_TYPE_P(z_fd)) {
-		case IS_RESOURCE:
-			if (!php_posix_stream_get_fd(z_fd, &fd)) {
-				RETURN_FALSE;
-			}
-			break;
-		default:
-			fd = zval_get_long(z_fd);
+	if (Z_TYPE_P(z_fd) == IS_RESOURCE) {
+		if (!php_posix_stream_get_fd(z_fd, &fd)) {
+			RETURN_FALSE;
+		}
+	} else {
+		if (!zend_parse_arg_long(z_fd, &fd, /* is_null */ NULL, /* check_null */ false, /* arg_num */ 1)) {
+			php_error_docref(NULL, E_WARNING, "Argument #1 ($file_descriptor) must be of type int|resource, %s given",
+				zend_zval_value_name(z_fd));
+			RETURN_FALSE;
+		}
 	}
 
+	/* A valid file descriptor must fit in an int and be positive */
+	if (fd < 0 || fd > INT_MAX) {
+		RETURN_FALSE;
+	}
 	if (isatty(fd)) {
 		RETURN_TRUE;
 	} else {
@@ -527,8 +545,8 @@ PHP_FUNCTION(posix_isatty)
 /* }}} */
 
 /*
-	POSIX.1, 4.8.1 sysconf() - TODO
-	POSIX.1, 5.7.1 pathconf(), fpathconf() - TODO
+	POSIX.1, 4.8.1 sysconf()
+	POSIX.1, 5.7.1 pathconf(), fpathconf()
 
 	POSIX.1, 5.1.2 opendir(), readdir(), rewinddir(), closedir()
 	POSIX.1, 5.2.1 chdir()
@@ -712,6 +730,44 @@ PHP_FUNCTION(posix_access)
 
 	RETURN_TRUE;
 }
+
+#ifdef HAVE_EACCESS
+PHP_FUNCTION(posix_eaccess)
+{
+	zend_long mode = 0;
+	size_t filename_len, ret;
+	char *filename, *path;
+
+	ZEND_PARSE_PARAMETERS_START(1, 2)
+		Z_PARAM_PATH(filename, filename_len)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_LONG(mode)
+	ZEND_PARSE_PARAMETERS_END();
+
+	path = expand_filepath(filename, NULL);
+	if (!path) {
+		zend_argument_value_error(1, "cannot be empty");
+		RETURN_THROWS();
+	}
+
+	if (php_check_open_basedir_ex(path, 0)) {
+		efree(path);
+		POSIX_G(last_error) = EPERM;
+		RETURN_FALSE;
+	}
+
+	ret = eaccess(path, mode);
+	efree(path);
+
+	if (ret) {
+		POSIX_G(last_error) = errno;
+		RETURN_FALSE;
+	}
+
+	RETURN_TRUE;
+}
+#endif
+
 /* }}} */
 
 /*
@@ -1088,16 +1144,43 @@ static const struct limitlist {
 PHP_FUNCTION(posix_getrlimit)
 {
 	const struct limitlist *l = NULL;
+	zend_long res;
+	bool res_is_null = true;
 
-	ZEND_PARSE_PARAMETERS_NONE();
+	ZEND_PARSE_PARAMETERS_START(0, 1)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_LONG_OR_NULL(res, res_is_null)
+	ZEND_PARSE_PARAMETERS_END();
 
-	array_init(return_value);
+	if (res_is_null) {
+		array_init(return_value);
 
-	for (l=limits; l->name; l++) {
-		if (posix_addlimit(l->limit, l->name, return_value) == FAILURE) {
-			zend_array_destroy(Z_ARR_P(return_value));
+		for (l=limits; l->name; l++) {
+			if (posix_addlimit(l->limit, l->name, return_value) == FAILURE) {
+				zend_array_destroy(Z_ARR_P(return_value));
+				RETURN_FALSE;
+			}
+		}
+	} else {
+		struct rlimit rl;
+		int result = getrlimit(res, &rl);
+		if (result < 0) {
+			POSIX_G(last_error) = errno;
 			RETURN_FALSE;
 		}
+
+		array_init(return_value);
+		if (rl.rlim_cur == RLIM_INFINITY) {
+			add_next_index_stringl(return_value, UNLIMITED_STRING, sizeof(UNLIMITED_STRING)-1);
+		} else {
+			add_next_index_long(return_value, rl.rlim_cur);
+		}
+
+		if (rl.rlim_max == RLIM_INFINITY) {
+			add_next_index_stringl(return_value, UNLIMITED_STRING, sizeof(UNLIMITED_STRING)-1);
+		} else {
+			add_next_index_long(return_value, rl.rlim_max);
+		}
 	}
 }
 /* }}} */
@@ -1177,3 +1260,79 @@ PHP_FUNCTION(posix_initgroups)
 }
 /* }}} */
 #endif
+
+PHP_FUNCTION(posix_sysconf)
+{
+	zend_long conf_id;
+
+	ZEND_PARSE_PARAMETERS_START(1, 1)
+		Z_PARAM_LONG(conf_id)
+	ZEND_PARSE_PARAMETERS_END();
+
+	RETURN_LONG(sysconf(conf_id));
+}
+
+#ifdef HAVE_PATHCONF
+PHP_FUNCTION(posix_pathconf)
+{
+	zend_long name, ret;
+	char *path;
+	size_t path_len;
+
+	ZEND_PARSE_PARAMETERS_START(2, 2)
+		Z_PARAM_PATH(path, path_len)
+		Z_PARAM_LONG(name);
+	ZEND_PARSE_PARAMETERS_END();
+
+	if (path_len == 0) {
+		zend_argument_value_error(1, "cannot be empty");
+		RETURN_THROWS();
+	} else if (php_check_open_basedir(path)) {
+		php_error_docref(NULL, E_WARNING, "Invalid path supplied: %s", path);
+		RETURN_FALSE;
+	}
+
+	ret = pathconf(path, name);
+
+	if (ret < 0 && errno != 0) {
+		POSIX_G(last_error) = errno;
+		RETURN_FALSE;
+	}
+
+	RETURN_LONG(ret);
+}
+#endif
+
+#ifdef HAVE_FPATHCONF
+PHP_FUNCTION(posix_fpathconf)
+{
+	zend_long name, ret, fd = 0;
+	zval *z_fd;
+
+	ZEND_PARSE_PARAMETERS_START(2, 2)
+		Z_PARAM_ZVAL(z_fd)
+		Z_PARAM_LONG(name);
+	ZEND_PARSE_PARAMETERS_END();
+
+	if (Z_TYPE_P(z_fd) == IS_RESOURCE) {
+		if (!php_posix_stream_get_fd(z_fd, &fd)) {
+			RETURN_FALSE;
+		}
+	} else {
+		if (!zend_parse_arg_long(z_fd, &fd, /* is_null */ NULL, /* check_null */ false, /* arg_num */ 1)) {
+			zend_argument_type_error(1, "must be of type int|resource, %s given",
+				zend_zval_value_name(z_fd));
+			RETURN_THROWS();
+		}
+	}
+
+	ret = fpathconf(fd, name);
+
+	if (ret < 0 && errno != 0) {
+		POSIX_G(last_error) = errno;
+		RETURN_FALSE;
+	}
+
+	RETURN_LONG(ret);
+}
+#endif
